/*

    File: exfat_adv.c

    Copyright (C) 2021 Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
 
#include <stdio.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <ctype.h>
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#include "types.h"
#include "common.h"
#include "intrf.h"
#include "intrfn.h"
#include "dirpart.h"
#include "exfat.h"
#include "exfat_adv.h"
#include "lang.h"
#include "io_redir.h"
#include "log.h"
#include "log_part.h"

static int log_exfat_info(const struct exfat_super_block *sb)
{
  const uint64_t start_exfat1=(uint64_t)le32(sb->fat_blocknr) << sb->blocksize_bits;
  log_info("start_sector %llu\n", (long long unsigned)le64(sb->start_sector));
  log_info("nr_sectors   %llu\n", (long long unsigned)le64(sb->nr_sectors));
  log_info("fat_blocknr  %u\n", le32(sb->fat_blocknr));
  log_info("fat_block_counts %u\n", le32(sb->fat_block_counts));
  log_info("clus_blocknr     %u\n", le32(sb->clus_blocknr));
  log_info("total_clusters   %u\n", le32(sb->total_clusters));
  log_info("rootdir_clusnr   %u\n", le32(sb->rootdir_clusnr));
  log_info("blocksize_bits   %u\n", sb->blocksize_bits);
  log_info("block_per_clus_bits %u\n", sb->block_per_clus_bits);
  log_info("number_of_fats   %u\n", sb->number_of_fats);

  log_info("start_exfat1 0x%llx\n", (long long unsigned)start_exfat1);
  return 0;
}

typedef struct exfat_offset_struct exfat_offset_t;

struct exfat_offset_struct
{
  unsigned long int offset;
  unsigned int  nbr;
};

static unsigned long int exfat_find_fat_start(const unsigned char *buffer, const unsigned int sector_size)
{
  /* TODO: handle limited size of exfat_offset */
  exfat_offset_t *exfat_offset;
  unsigned int nbr_offset=0;
  int have_fat_signature=0;
  exfat_offset=(exfat_offset_t *)MALLOC(sector_size*sizeof(exfat_offset_t));
  {
    unsigned int i,j;
    const uint32_t *p32=(const uint32_t*)buffer;
    unsigned int err=0;
    for(i=0; (i<sector_size/4)&&(err==0); i++)
    {
      unsigned long int cluster=le32(p32[i])&0x0FFFFFFF;
      if(cluster==1)
      {
	err=1;
      }
      if((cluster!=0) && ((cluster&0x0ffffff8)!=(unsigned)0x0ffffff8))
      {
	for(j=i+1; (j<sector_size/4)&&(err==0); j++)
	{
	  if(cluster==(le32(p32[j])&0x0FFFFFFF))
	  {
	    err=1;
	  }
	}
      }
    }
    if(err==0)
    {
      for(i=0; i<sector_size/4; i++)
      {
	unsigned long int cluster=le32(p32[i])&0x0FFFFFFF;
	if((cluster!=0) && ((cluster&0x0ffffff8)!=(unsigned)0x0ffffff8)&&((cluster-i-1)%(sector_size/4)==0))
	{
	  for(j=0;(j<nbr_offset) &&
	      (exfat_offset[j].offset!=(cluster-i-1)/(sector_size/4));j++);
	  if(j<nbr_offset)
	    exfat_offset[j].nbr++;
	  else
	  {
	    exfat_offset[nbr_offset].offset=(cluster-i-1)/(sector_size/4);
	    exfat_offset[nbr_offset].nbr=1;
	    nbr_offset++;
	  }
	}
      }
    }
    if((buffer[0]==0xF0 || buffer[0]>=0xF8) && buffer[1]==0xFF &&
	buffer[2]==0xFF && ((buffer[3]==0x0F) ||(buffer[3]==0xFF)) &&
	buffer[4]==0xFF && buffer[5]==0xFF && buffer[6]==0xFF)
    {
      for(j=0; j<nbr_offset && exfat_offset[j].offset!=0; j++);
      if(j<nbr_offset)
	exfat_offset[j].nbr++;
      else
      {
	exfat_offset[nbr_offset].offset=0;
	exfat_offset[nbr_offset].nbr=1;
	nbr_offset++;
      }
      have_fat_signature=1;
    }
  }
  if(nbr_offset>0)
  {
    unsigned int j;
    unsigned int best_j=0;
    for(j=0;j<nbr_offset;j++)
    {
      if(exfat_offset[j].nbr>exfat_offset[best_j].nbr)
	best_j=j;
    }
    if(exfat_offset[best_j].nbr>10 || have_fat_signature>0)
    {
      const unsigned long int res=exfat_offset[best_j].offset;
      free(exfat_offset);
      return res;
    }
  }
  free(exfat_offset);
  return 0;
}

static int exfat_find_fat(disk_t *disk, const partition_t *partition, const uint64_t max_offset, unsigned int *nbr_offset, exfat_offset_t *exfat_offset, const unsigned int max_nbr_offset)
{
  uint64_t offset;
#ifdef HAVE_NCURSES
  unsigned long int old_percent=0;
#endif
  int ind_stop=0;
  unsigned char *buffer=(unsigned char *)MALLOC(disk->sector_size);
#ifdef HAVE_NCURSES
  wmove(stdscr,22,0);
  wattrset(stdscr, A_REVERSE);
  waddstr(stdscr,"  Stop  ");
  wattroff(stdscr, A_REVERSE);
#endif
  for(offset=disk->sector_size;
      offset<max_offset && !ind_stop;
      offset+=disk->sector_size)
  {
#ifdef HAVE_NCURSES
    const unsigned long int percent=offset*100/max_offset;
    if(percent!=old_percent)
    {
      wmove(stdscr,8,30);
      wclrtoeol(stdscr);	/* before addstr for BSD compatibility */
      wprintw(stdscr,"Searching for FAT table %lu%%",percent);
      old_percent=percent;
      wrefresh(stdscr);
      ind_stop|=check_enter_key_or_s(stdscr);
    }
#endif
    if((unsigned)disk->pread(disk, buffer, disk->sector_size, partition->part_offset + offset) == disk->sector_size)
    {
      const unsigned long int fat_offset=exfat_find_fat_start(buffer, disk->sector_size);
      if(fat_offset > 0 && fat_offset<=(offset/disk->sector_size))
      {
	unsigned int j;
	for(j=0; j<*nbr_offset && exfat_offset[j].offset!=offset/disk->sector_size-fat_offset; j++);
	if(j<*nbr_offset)
	  exfat_offset[j].nbr++;
	else
	{
	  unsigned int new_info=0;
	  if(*nbr_offset<max_nbr_offset)
	  {
	    new_info=*nbr_offset;
	    (*nbr_offset)++;
	  }
	  else
	  { /* Overwrite the last information field with the lower nbr */
	    for(j=0;j<max_nbr_offset;j++)
	    {
	      if(exfat_offset[j].nbr <= exfat_offset[new_info].nbr)
		new_info=j;
	    }
	  }
	  exfat_offset[new_info].offset=offset/disk->sector_size-fat_offset;
	  exfat_offset[new_info].nbr=1;
	}
      }
    }
  }
#ifdef HAVE_NCURSES
  wmove(stdscr,22,0);
  wclrtoeol(stdscr);
  wrefresh(stdscr);
#endif
  free(buffer);
  return 0;
}

int rebuild_exfat_BS(disk_t *disk, partition_t *partition, char **current_cmd)
{
  unsigned long int fat_length_max;
  unsigned long int max_offset;
  unsigned int nbr_offset=0;
  unsigned int i;
  exfat_offset_t info_offset[0x400];
  log_info("rebuild_exfat_BS\n");
  fat_length_max=partition->part_size/disk->sector_size*4;
  max_offset=fat_length_max+64*disk->sector_size;
  max_offset=partition->part_size;
  {
    char buffer[512];
    if(disk->pread(disk, buffer, sizeof(buffer), partition->part_offset) == sizeof(buffer))
    {
      const struct exfat_super_block *sb=(const struct exfat_super_block *)buffer;
      log_exfat_info(sb);
    }
  }
  exfat_find_fat(disk, partition, max_offset, &nbr_offset, &info_offset[0], 0x400);
  for(i=0;i<nbr_offset;i++)
  {
    log_info("exfat at %lu, nbr=%u\n", info_offset[i].offset, info_offset[i].nbr);
  }
  return 0;
}
